unit ClipperRoutines;

{$mode delphi}

{
2015-07-10 Unit med ett antal procedurer och funktioner för klippning
           som utnyttjar Angus Johnsons superba Clipper
}

interface

uses
  LCLIntf, LCLType, LMessages, Classes, SysUtils, StrUtils, Clipper,
  ShapeObj, FileUtil, IniFiles, Dialogs, ComCtrls, Forms;

type

  { TClipperRoutines }

  TClipperRoutines = class(TObject)
  private
    { Private declarations }

  public
    ZFillCallbackProcedure : TZFillCallback;
    constructor Create(AOwner: TObject);
    destructor Destroy; override;

   procedure AddPath(var paths: TPaths; path: TPath); //Closed: Boolean);
   procedure BufferPaths(paths: TPaths; inWards: boolean;
     bufferWidth : single; clipperScaleFactor : integer;
     var bufferedPaths: TPaths);
   function BufferShape(shape: TShapeObj; inWards: boolean;
     bufferWidth : single; var bufferedShape: TShapeObj) : string;
   procedure ClearPaths(var paths: TPaths);
   procedure ClipPaths(subjectPaths, clipPaths: TPaths; clipType: TClipType;
     var clippedPaths: TPaths; ProgressBar: TProgressBar);
   procedure CopyPaths(copyFromPaths: TPaths; var copyToPaths: TPaths);
   procedure LoadDataMillAsPaths(dataMillFileName: string;
     clipperScaleFactor: integer; var dataMillPaths: TPaths);
   procedure LoadModelRectangleAsPaths(modelName: string;
     clipperScaleFactor: integer; var paths: TPaths);
   function LoadPathFromShape(var path: TPath;
     var pathXMin, pathYMin, pathXMax, pathYMax : integer;
     shape: TShapeObj;
     shapeNum : integer; xorigo, yorigo, xyScale, zScale : single;
     clipperScaleFactor : integer) : string;
   procedure LoadPathsFromShape(var paths: TPaths; shape: TShapeObj;
     clipperScaleFactor : integer); overload;
   function LoadPathsFromShape(var paths: TPaths; shape: TShapeObj;
     xorigo, yorigo, xyScale, zScale : single;
     clipperScaleFactor : integer) : string; overload;
   function LoadPathsFromShape(var paths: TPaths; shape: TShapeObj;
     shapeNum : integer; xorigo, yorigo, xyScale, zScale : single;
     clipperScaleFactor : integer) : string; overload;
   function LoadShapeFromPaths(var shape: TShapeObj; paths: TPaths;
     clipperScaleFactor : integer) : string; overload;
   function LoadShapeFromPaths(var shape: TShapeObj; paths: TPaths;
     xorigo, yorigo, xyScale, zScale : single;
     clipperScaleFactor : integer) : string; overload;
   function LoadShapeFromPath(var shape: TShapeObj; path: TPath;
     xorigo, yorigo, xyScale, zScale : single;
     clipperScaleFactor : integer) : string; overload;
   procedure WritePath(path: TPath; fileName: string;
     clipperScaleFactor : integer);
   procedure WritePaths(paths: TPaths; fileName: string;
     clipperScaleFactor : integer);
  end;

implementation

{ TClipperRoutines }

constructor TClipperRoutines.Create(AOwner: TObject);
begin
//  inherited Create(AOwner);
end;

destructor TClipperRoutines.Destroy;
begin
  inherited Destroy;
end;

procedure ZFillCallbackProcedure(
  const E1Bot, E1Top, E2Bot, E2Top: TIntPoint; var Pt: TIntPoint);
begin
  ShowMessage('Hej ' +
  'X:' + FloatToStrF((1.0 - 1.0 + E1Bot.X) / 100000, ffFixed, 15, 2) + ' ' +
  'Y:' + FloatToStrF((1.0 - 1.0 + E1Bot.Y) / 100000, ffFixed, 15, 2) + ' ' +
  'Z:' + FloatToStrF((1.0 - 1.0 + E1Bot.Z) / 100000, ffFixed, 15, 2));
end;

procedure TClipperRoutines.BufferPaths(paths : TPaths; inWards : boolean;
  bufferWidth : single; clipperScaleFactor : integer; var bufferedPaths : TPaths);
var
  i, n, l : integer;
  tempPaths, closedPaths, openPaths : TPaths;
  orientering : boolean;

{  procedure SeparateClosedAndOpenPaths(paths : TPaths;
    var closedPaths, openPaths : TPaths);
  var
    n, nc, no : integer;
  begin
    nc := 0;
    no := 0;
    n := 0;
    while n < Length(paths) do
    begin
      if (paths[n][0].X = paths[n][Length(paths[n]) - 1].X) and
         (paths[n][0].Y = paths[n][Length(paths[n]) - 1].Y) and
         (paths[n][0].Z = paths[n][Length(paths[n]) - 1].Z) then
      begin
        SetLength(closedPaths, nc + 1);
        closedPaths[nc] := paths[n];
        Inc(nc);
      end
      else
      begin
        SetLength(openPaths, no + 1);
        openPaths[no] := paths[n];
        Inc(no);
      end;

      Inc(n);
    end;
  end;

  procedure BufferClosedPathsBufferPaths(closedPaths : TPaths; inWards : boolean;
    bufferWidth : single; scaleFactor : integer;
    var bufferedPaths : TPaths);
  var
    tempPaths : TPaths;
  begin
    with TClipperOffset.Create do
    try
      AddPaths(closedPaths, jtRound, etClosedPolygon);//etOpenButt);
      Execute(tempPaths, bufferWidth * scaleFactor);
    finally
      Free;
    end;

    n := Length(tempPaths);
    i := 0;
    n := 0;
    while n < Length(tempPaths) do
    begin
      if (not inWards) and (Orientation(closedPaths[n])) or
         (inWards) and (not Orientation(closedPaths[n])) then
      begin
        SetLength(bufferedPaths, i + 1);
        bufferedPaths[i] := tempPaths[n];

        if (closedPaths[n][0].X = paths[n][Length(closedPaths[n]) - 1].X) and
           (closedPaths[n][0].Y = paths[n][Length(closedPaths[n]) - 1].Y) and
           (closedPaths[n][0].Z = paths[n][Length(closedPaths[n]) - 1].Z) then
        begin
          l := Length(bufferedPaths[i]);
          SetLength(bufferedPaths[i], l + 1);
          bufferedPaths[i][Length(bufferedPaths[i]) - 1].X :=
            bufferedPaths[i][0].X;
          bufferedPaths[i][Length(bufferedPaths[i]) - 1].Y :=
            bufferedPaths[i][0].Y;
          bufferedPaths[i][Length(bufferedPaths[i]) - 1].Z :=
            bufferedPaths[i][0].Z;
        end;

        Inc(i);
      end;

      Inc(n);
    end;
  end;

  procedure BufferOpenPathsBufferPaths(openPaths : TPaths; inWards : boolean;
    bufferWidth : single; scaleFactor : integer;
    var bufferedPaths : TPaths);
  var
    tempPaths : TPaths;
  begin
    with TClipperOffset.Create do
    try
      AddPaths(openPaths, jtRound, etOpenButt);
      Execute(tempPaths, abs(bufferWidth * scaleFactor));
    finally
      Free;
    end;

    n := Length(tempPaths);
    i := 0;
    n := 0;
    while n < Length(tempPaths) do
    begin
      if (not inWards) and (Orientation(openPaths[n])) or
         (inWards) and (not Orientation(openPaths[n])) then
      begin
        SetLength(bufferedPaths, i + 1);
        bufferedPaths[i] := tempPaths[n];

        Inc(i);
      end;

      Inc(n);
    end;
  end;}

begin
{  ClearPaths(closedPaths);
  ClearPaths(openPaths);
  ClearPaths(bufferedPaths);

  SeparateClosedAndOpenPaths(paths, closedPaths, openPaths);
  BufferClosedPathsBufferPaths(closedPaths, inWards, bufferWidth, scaleFactor,
    bufferedPaths);
  BufferOpenPathsBufferPaths(openPaths, inWards, bufferWidth, scaleFactor,
    bufferedPaths);}

  ClearPaths(bufferedPaths);

  with TClipperOffset.Create do
  try
    ArcTolerance := 0.1 * clipperScaleFactor;
    AddPaths(paths, jtRound, etClosedPolygon);//etOpenButt);
    if inWards then
      Execute(tempPaths, -bufferWidth * clipperScaleFactor)
    else
      Execute(tempPaths, bufferWidth * clipperScaleFactor);
  finally
    Free;
  end;

  // stäng ev. öppna polygoner
  l := Length(tempPaths);
  n := 0;
  while n < Length(tempPaths) do
  begin
    SetLength(bufferedPaths, n + 1);
    bufferedPaths[n] := tempPaths[n];

    if (paths[n][0].X = paths[n][Length(paths[n]) - 1].X) and
       (paths[n][0].Y = paths[n][Length(paths[n]) - 1].Y) then
    begin
      // stäng öppen polygon
      SetLength(bufferedPaths[n], Length(bufferedPaths[n]) + 1);
      bufferedPaths[n][Length(bufferedPaths[n]) - 1].X :=
        bufferedPaths[n][0].X;
      bufferedPaths[n][Length(bufferedPaths[n]) - 1].Y :=
        bufferedPaths[n][0].Y;
    end;

    Inc(n);
  end;
end;

function TClipperRoutines.BufferShape(shape: TShapeObj; inWards: boolean;
  bufferWidth : single; var bufferedShape: TShapeObj) : string;
var
  paths, bufferedPaths : TPaths;
begin
  LoadPathsFromShape(paths, shape, 10000000);
  BufferPaths(paths, inWards, bufferWidth, 10000000, bufferedPaths);
  Result := LoadShapeFromPaths(bufferedShape, bufferedPaths, 10000000);
end;

procedure TClipperRoutines.AddPath(var paths : TPaths; path : TPath);
//  Closed: Boolean);
var
  i : integer;
begin
  SetLength(paths, Length(paths) + 1);
  SetLength(paths[Length(paths) - 1], Length(path));
  i := 0;
  while i < Length(path) do
  begin
    paths[Length(paths) - 1][i].X := path[i].X;
    paths[Length(paths) - 1][i].Y := path[i].Y;
    paths[Length(paths) - 1][i].Z := path[i].Z;

    Inc(i);
  end;
end;

procedure TClipperRoutines.ClearPaths(var paths : TPaths);
var
  i : integer;
begin
  i := 0;
  while i < Length(paths) do
  begin
    SetLength(paths[i], 0);

    Inc(i);
  end;
  SetLength(paths, 0);
end;

procedure TClipperRoutines.ClipPaths(subjectPaths, clipPaths : TPaths;
  clipType : TClipType; var clippedPaths : TPaths; ProgressBar: TProgressBar);
var
  polyTree : TPolyTree;
  paths : TPaths;
  res : boolean;
  clipper : TClipper;
  n : integer;
begin
  ClearPaths(paths);
  ClearPaths(clippedPaths);

  // för att få klippta polylines/polygoner sorterade i samma ordning
  // motsv. oklippta sker klippning av dessa en och en som sparas i
  // resulterande clippedPaths
  polyTree := TPolyTree.Create;
  clipper := TClipper.Create;
  try
{    clipper := TClipper.Create;
    try}
//      clipper.ReverseSolution := true;
      clipper.LinearZInterpolation := true;
      if ProgressBar <> nil then
      begin
        ProgressBar.Max := Length(subjectPaths);
        ProgressBar.Min := 0;
        ProgressBar.Position := 0;
      end;

      n := Length(subjectPaths);
      n := 0;
      while n < Length(subjectPaths) do
      begin
        polyTree.Clear;
        clipper.Clear;
        clipper.AddPath(subjectPaths[n], ptSubject, false);
        clipper.AddPaths(clipPaths, ptClip, true);
        res := clipper.Execute(clipType, polyTree, pftNonZero, pftNonZero);
        paths := OpenPathsFromPolyTree(polyTree);
        if Length(paths) > 0 then
          AddPath(clippedPaths, paths[0]);

        Inc(n);
        if (ProgressBar <> nil) and (n mod 10 = 0) then
        begin
          Application.ProcessMessages;
          ProgressBar.Position := n;
          ProgressBar.Refresh;
          Sleep(10);
        end;
      end;

      if ProgressBar <> nil then
        ProgressBar.Position := 0;

      clippedPaths := ReversePaths(clippedPaths);
{    finally
      FreeAndNil(clipper);
    end;

    clippedPaths := OpenPathsFromPolyTree(polyTree);}
  finally
    FreeAndNil(clipper);
    FreeAndNil(polyTree);
  end;
end;

procedure TClipperRoutines.CopyPaths(copyFromPaths : TPaths; var copyToPaths : TPaths);
var
  i, n : integer;
begin
  ClearPaths(copyToPaths);

  SetLength(copyToPaths, Length(copyFromPaths));
  i := 0;
  while i < Length(copyFromPaths) do
  begin
    SetLength(copyToPaths[i], Length(copyFromPaths[i]));
    n := 0;
    while n < Length(copyFromPaths[i]) do
    begin
      copyToPaths[i][n].X := copyFromPaths[i][n].X;
      copyToPaths[i][n].Y := copyFromPaths[i][n].Y;
      copyToPaths[i][n].Z := copyFromPaths[i][n].Z;

      Inc(n);
    end;
    Inc(i);
  end;
end;

procedure TClipperRoutines.LoadDataMillAsPaths(dataMillFileName : string;
  clipperScaleFactor : integer; var dataMillPaths : TPaths);
var
  i, n : integer;
  dataShapeName : string;
  dataMillShapeObj : TShapeObj;
begin
  dataMillShapeObj := TShapeObj.Create(nil);
  try
    dataMillShapeObj.Open(dataMillFileName, 'ro');

    ClearPaths(dataMillPaths);

    // ladda in alla shapes som inte är markerade som raderade
    i := 0;
    while i < dataMillShapeObj.NumEntities do
    begin
      dataMillShapeObj.ReadShape(i);
      if dataMillShapeObj.SingleInstanceShapeType <> 0 then
      begin
        setlength(dataMillPaths, Length(dataMillPaths) + 1);
        setlength(dataMillPaths[Length(dataMillPaths) - 1],
          dataMillShapeObj.NumPoints);
        n := 0;
        while n < dataMillShapeObj.NumPoints do
        begin
          dataMillPaths[i][n].X :=
            Round(dataMillShapeObj.XYArray[n].x * clipperScaleFactor);
          dataMillPaths[i][n].Y :=
            Round(dataMillShapeObj.XYArray[n].y * clipperScaleFactor);

          Inc(n);
        end;
      end;

      Inc(i);
    end;
  finally
    FreeAndNil(dataMillShapeObj);
  end;
end;

procedure TClipperRoutines.LoadModelRectangleAsPaths(
  modelName : string; clipperScaleFactor : integer; var paths : TPaths);
var
  i : integer;
  iniFile : TIniFile;
  xmax, ymax : single;
  bskFileName : string;
begin
  if PosEx('.', modelName) <= 0 then
    bskFileName := modelName + '.bsk'
  else
    bskFileName := Copy(modelName, 1, PosEx('.', modelName) - 1) + '.bsk';
  iniFile := TIniFile.Create(bskFileName);
  try
    xmax := iniFile.ReadFloat('Fräsning',
      'Storlek i X-led', 0);
    ymax := iniFile.ReadFloat('Fräsning',
      'Storlek i Y-led', 0);
{    xmax := iniFile.ReadFloat(UTF8ToSys('Fräsning'),
      'Storlek i X-led', 0);
    ymax := iniFile.ReadFloat(UTF8ToSys('Fräsning'),
      'Storlek i Y-led', 0);}

    ClearPaths(paths);

    setlength(paths, 1);
    setlength(paths[0], 5);
    paths[0][0].X := 0;
    paths[0][0].Y := 0;
    paths[0][1].X := 0;
    paths[0][1].Y := Round(ymax * clipperScaleFactor);
    paths[0][2].X := Round(xmax * clipperScaleFactor);
    paths[0][2].Y := Round(ymax * clipperScaleFactor);
    paths[0][3].X := Round(xmax * clipperScaleFactor);
    paths[0][3].Y := 0;
    paths[0][4].X := 0;
    paths[0][4].Y := 0;
  finally
    FreeAndNil(iniFile);
  end;
end;

function TClipperRoutines.LoadPathFromShape(var path: TPath;
  var pathXMin, pathYMin, pathXMax, pathYMax : integer;
  shape: TShapeObj;
  shapeNum: integer; xorigo, yorigo, xyScale, zScale: single;
  clipperScaleFactor: integer): string;
var
  n : integer;
begin
  SetLength(path, 0);

  pathXMin := 100000000;
  pathYMin := 100000000;
  pathXMax := -100000000;
  pathYMax := -100000000;

  shape.ReadShape(shapeNum);
  if shape.SingleInstanceShapeType <> 0 then
  begin
    setlength(path, shape.NumPoints);
    n := 0;
    while n < shape.NumPoints do
    begin
      path[n].X :=
        Round(((shape.XYArray[n].x - xorigo) / xyScale) * clipperScaleFactor);
      path[n].Y :=
        Round(((shape.XYArray[n].y - yorigo) / xyScale) * clipperScaleFactor);
      if Length(shape.ZArray) > 0 then
        path[n].Z :=
            Round(((shape.ZArray[n] / xyScale) / ZScale) * clipperScaleFactor);

      if pathXMin > path[n].X then
        pathXMin := path[n].X;
      if pathYMin > path[n].Y then
        pathYMin := path[n].Y;
      if pathXMax < path[n].X then
        pathXMax := path[n].X;
      if pathYMax < path[n].Y then
        pathYMax := path[n].Y;

      Inc(n);
    end;
  end;
end;

procedure TClipperRoutines.LoadPathsFromShape(var paths : TPaths;
  shape : TShapeObj; clipperScaleFactor : integer);
var
  i, n : integer;
begin
  ClearPaths(paths);

  // ladda in alla shapes som inte är markerade som raderade
  i := 0;
  while i < shape.NumEntities do
  begin
    shape.ReadShape(i);
    if shape.SingleInstanceShapeType <> 0 then
    begin
      setlength(paths, Length(paths) + 1);
      setlength(paths[Length(paths) - 1], shape.NumPoints);
      n := 0;
      while n < shape.NumPoints do
      begin
        paths[Length(paths) - 1][n].X :=
          Round(shape.XYArray[n].y * clipperScaleFactor);
        paths[Length(paths) - 1][n].Y :=
          Round(shape.XYArray[n].x * clipperScaleFactor);
        if Length(shape.ZArray) > 0 then
          paths[Length(paths) - 1][n].Z :=
              Round(shape.ZArray[n] * clipperScaleFactor);

        Inc(n);
      end;
    end;

    Inc(i);
  end;
end;

function TClipperRoutines.LoadPathsFromShape(var paths: TPaths;
  shape: TShapeObj; xorigo, yorigo, xyScale, zScale: single;
  clipperScaleFactor: integer): string;
var
  shapeNum, n : integer;
begin
  ClearPaths(paths);

  // ladda in alla shapes som inte är markerade som raderade
  shapeNum := 0;
  while shapeNum < shape.NumEntities do
  begin
    shape.ReadShape(shapeNum);
    if shape.SingleInstanceShapeType <> 0 then
    begin
      setlength(paths, Length(paths) + 1);
      setlength(paths[Length(paths) - 1], shape.NumPoints);
      n := 0;
      while n < shape.NumPoints do
      begin
        paths[Length(paths) - 1][n].X :=
          Round((shape.XYArray[n].x - xorigo) / xyScale) * clipperScaleFactor;
        paths[Length(paths) - 1][n].Y :=
          Round((shape.XYArray[n].y - yorigo) / xyScale) * clipperScaleFactor;
        if Length(shape.ZArray) > 0 then
          paths[Length(paths) - 1][n].Z :=
              Round((shape.ZArray[n] / xyScale) / ZScale) * clipperScaleFactor;

        Inc(n);
      end;
    end;

    Inc(shapeNum);
  end;
end;

function TClipperRoutines.LoadPathsFromShape(var paths: TPaths;
  shape: TShapeObj; shapeNum: integer; xorigo, yorigo, xyScale, zScale: single;
  clipperScaleFactor: integer): string;
var
  n : integer;
begin
  ClearPaths(paths);

  shape.ReadShape(shapeNum);
  if shape.SingleInstanceShapeType <> 0 then
  begin
    setlength(paths, Length(paths) + 1);
    setlength(paths[Length(paths) - 1], shape.NumPoints);
    n := 0;
    while n < shape.NumPoints do
    begin
      paths[Length(paths) - 1][n].X :=
        Round(((shape.XYArray[n].x - xorigo) / xyScale) * clipperScaleFactor);
      paths[Length(paths) - 1][n].Y :=
        Round(((shape.XYArray[n].y - yorigo) / xyScale) * clipperScaleFactor);
      if Length(shape.ZArray) > 0 then
        paths[Length(paths) - 1][n].Z :=
            Round(((shape.ZArray[n] / xyScale) / ZScale) * clipperScaleFactor);

      Inc(n);
    end;
  end;
end;

function TClipperRoutines.LoadShapeFromPaths(var shape: TShapeObj; paths: TPaths;
  clipperScaleFactor: integer) : string;
var
  i, n : integer;
  x, y, z : single;
begin
  Result := 'Akt. shape inte i "rw"-mode';

  if shape.Mode <> 'rw' then Exit;

  i := 0;
  while i < Length(paths) do
  begin
    shape.AddPolygonPolylineShape;

    n := 0;
    while n < Length(paths[i]) do
    begin
      x := (1.0 - 1.0 + paths[i][n].Y) / clipperScaleFactor;
      y := (1.0 - 1.0 + paths[i][n].X) / clipperScaleFactor;
      z := (1.0 - 1.0 + paths[i][n].Z) / clipperScaleFactor;
      shape.Add(x, y, z, 0);

      Inc(n);
    end;

    Inc(i);
  end;
  shape.ClosePolygonPolylineShape;

  Result := 'OK';
end;

function TClipperRoutines.LoadShapeFromPaths(var shape: TShapeObj;
  paths: TPaths; xorigo, yorigo, xyScale, zScale: single;
  clipperScaleFactor: integer): string;
var
  i, n : integer;
  x, y, z : single;
begin
  Result := 'Akt. shape inte i "rw"-mode';

  if shape.Mode <> 'rw' then Exit;

  i := 0;
  while i < Length(paths) do
  begin
    LoadShapeFromPath(shape, paths[i], xorigo, yorigo, xyScale, zScale,
      clipperScaleFactor);

    Inc(i);
  end;

  Result := 'OK';
end;

function TClipperRoutines.LoadShapeFromPath(var shape: TShapeObj;
  path: TPath; xorigo, yorigo, xyScale, zScale: single;
  clipperScaleFactor: integer): string;
var
  i, n : integer;
  x, y, z : single;
begin
  Result := 'Akt. shape inte i "rw"-mode';

  if shape.Mode <> 'rw' then Exit;

  shape.AddPolygonPolylineShape;

  n := 0;
  while n < Length(path) do
  begin
    x := xorigo +
      ((1.0 - 1.0 + path[n].X) / clipperScaleFactor) * xyScale;
    y := yorigo +
      ((1.0 - 1.0 + path[n].Y) / clipperScaleFactor) * xyScale;
    z := ((1.0 - 1.0 + path[n].Z) / clipperScaleFactor) *
      xyScale * zScale;
    shape.Add(x, y, z, 0);

    Inc(n);
  end;
  shape.ClosePolygonPolylineShape;

  Result := 'OK';
end;

procedure TClipperRoutines.WritePath(path: TPath; fileName: string;
  clipperScaleFactor: integer);
var
  i, n : integer;
  x, y, z : single;
  tempList : TStringList;
begin
  tempList := TStringList.Create;
  try
    tempList.Add('Path: ' + IntToStr(i) +
      ' NumPoints: ' + IntToStr(Length(path)));
    n := 0;
    while n < Length(path) do
    begin
      x := path[n].X;
      y := path[n].Y;
      z := path[n].Z;
      tempList.Add(
        FloatToStrF((1.0 - 1.0 + x) / clipperScaleFactor, ffFixed, 15, 4) + ' ' +
        FloatToStrF((1.0 - 1.0 + y) / clipperScaleFactor, ffFixed, 15, 4) + ' ' +
        FloatToStrF((1.0 - 1.0 + z) / clipperScaleFactor, ffFixed, 15, 4));

      Inc(n);
    end;
  finally
    repeat
      fileName := AnsiReplaceText(fileName, '.txt', '');
    until PosEx('.txt', fileName) <= 0;
    tempList.SaveToFile(fileName + '.txt');
//    tempList.SaveToFile(UTF8ToSys(fileName + '.txt'));
    FreeAndNil(tempList);
  end;
end;

procedure TClipperRoutines.WritePaths(paths : TPaths; fileName : string;
  clipperScaleFactor : integer);
var
  i, n : integer;
  x, y, z : single;
  tempList : TStringList;
begin
  tempList := TStringList.Create;
  try
    tempList.Add('NumPaths: ' + IntToStr(Length(paths)));
    i := 0;
    while i < Length(paths) do
    begin
      tempList.Add('Path: ' + IntToStr(i) +
        ' NumPoints: ' + IntToStr(Length(paths[i])));
      n := 0;
      while n < Length(paths[i]) do
      begin
        x := paths[i][n].X;
        y := paths[i][n].Y;
        z := paths[i][n].Z;
        tempList.Add(
          FloatToStrF((1.0 - 1.0 + x) / clipperScaleFactor, ffFixed, 15, 4) + ' ' +
          FloatToStrF((1.0 - 1.0 + y) / clipperScaleFactor, ffFixed, 15, 4) + ' ' +
          FloatToStrF((1.0 - 1.0 + z) / clipperScaleFactor, ffFixed, 15, 4));

        Inc(n);
      end;

      Inc(i);
    end;
  finally
    repeat
      fileName := AnsiReplaceText(fileName, '.txt', '');
    until PosEx('.txt', fileName) <= 0;
    tempList.SaveToFile(fileName + '.txt');
//    tempList.SaveToFile(UTF8ToSys(fileName + '.txt'));
    FreeAndNil(tempList);
  end;
end;

end.

