unit ShapeObjDemomain;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, Menus, ComCtrls,
  ExtCtrls, ShapeObj, WorldViewPortRoutines, WGS84, ClipperRoutines,
  StrUtils;

type

  { TfrmShapeObjDemo }

  TfrmShapeObjDemo = class(TForm)
    chkgrp1: TCheckGroup;
    lbledPenWidth: TLabeledEdit;
    MainMenu1: TMainMenu;
    itemFile: TMenuItem;
    itemImportXYZTextData: TMenuItem;
    itemExport: TMenuItem;
    itemSimpleXYZFile: TMenuItem;
    itemEntityFormat: TMenuItem;
    itemShowShapeFiles: TMenuItem;
    itemAttributes: TMenuItem;
    itemExit: TMenuItem;
    itemTools: TMenuItem;
    itemDeleteShapeFiles: TMenuItem;
    itemCopyShapeFiles: TMenuItem;
    itemMergeShapeFiles: TMenuItem;
    OpenDialog1: TOpenDialog;
    PaintBox1: TPaintBox;
    Panel1: TPanel;
    ProgressBar1: TProgressBar;
    rgpShapeType: TRadioGroup;
    StatusBar1: TStatusBar;
    procedure chkgrp1ItemClick(Sender: TObject; Index: integer);
    procedure FormCreate(Sender: TObject);
    procedure itemAttributesClick(Sender: TObject);
    procedure itemCopyShapeFilesClick(Sender: TObject);
    procedure itemDeleteShapeFilesClick(Sender: TObject);
    procedure itemEntityFormatClick(Sender: TObject);
    procedure itemExitClick(Sender: TObject);
    procedure itemImportXYZTextDataClick(Sender: TObject);
    procedure itemShowShapeFilesClick(Sender: TObject);
    procedure itemSimpleXYZFileClick(Sender: TObject);
    procedure lbledPenWidthChange(Sender: TObject);
    procedure PaintBox1MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
  private
    FShapeObj : TShapeObj;
    FWorldViewPortRoutines : TWorldViewPortRoutines;
    FPenWidth : integer;
    FShapefileList : TStringList;
    procedure DrawShapes;
  public

  end;

var
  frmShapeObjDemo: TfrmShapeObjDemo;

implementation

{$R *.lfm}

{ TfrmShapeObjDemo }

procedure TfrmShapeObjDemo.FormCreate(Sender: TObject);
begin
  FShapeObj := TShapeObj.Create(Self);
  FWorldViewPortRoutines := TWorldViewPortRoutines.Create(Self);
  FShapeObj.WorldViewPortRoutines := FWorldViewPortRoutines;
  FShapefileList := TStringList.Create;
  FPenWidth := 1;
end;

procedure TfrmShapeObjDemo.chkgrp1ItemClick(Sender: TObject; Index: integer);
begin
  DrawShapes;
end;

procedure TfrmShapeObjDemo.itemExitClick(Sender: TObject);
begin
  FreeAndNil(FShapefileList);
  Close;
end;

procedure TfrmShapeObjDemo.itemImportXYZTextDataClick(Sender: TObject);
var
  shapeFileName, line, res : string;
  tempList, tempList2 : TStringList;
  n, numValues, numPoints : integer;
  pointDataType : boolean;
  x, y, z : single;
  shapeType : integer;

  function ExistsForPossibleOldShapeFileWithSameName : boolean;
  begin
    Result := true;
    // check for possible already existing shapefile with current name
    if FShapeObj.Exists(shapeFileName) then
    begin
      if MessageDlg('', ExtractFilename(shapeFileName) +
        ' already exist' + lineEnding + 'OK to overwrite', mtConfirmation,
        [mbOK, mbNo], 0, mbNo) = mrNo then Exit;

      res := FShapeObj.DeleteShapeFile(shapeFileName);
      if res <> 'Done' then
      begin
        ShowMessage('Did not succeed to delete old shapefile:' + lineEnding +
          shapeFileName + lineEnding +
          'maybe open in some other program');
        Exit;
      end;
    end;

    Result := false;
  end;

  procedure RemovePossibleTabsAndDoubleSpaces;
  begin
    repeat
      line := AnsiReplaceText(line, chr(9) + chr(9), ' ');
    until PosEx(chr(9) + chr(9), line) <= 0;
    repeat
      line := AnsiReplaceText(line, chr(9), ' ');
    until PosEx(chr(9), line) <= 0;
    repeat
      line := AnsiReplaceText(line, '  ', ' ');
    until PosEx('  ', line) <= 0;
    line := TrimRight(TrimLeft(line));
  end;

  procedure CreateShapeFile;
  var
    n : integer;

    function GetNumberOfValues : integer;
    begin
      line := tempList.Strings[0];
      RemovePossibleTabsAndDoubleSpaces;
      tempList2.Text := AnsiReplaceText(line, ' ', lineEnding);
      Result := tempList2.Count;
    end;

  begin
    numValues := GetNumberOfValues;

    case rgpShapeType.ItemIndex of
      0:
      begin
        if numValues = 2 then
          shapeType := 1
        else
          shapeType := 11;
      end;
      1:
      begin
        if numValues = 2 then
          shapeType := 3
        else
          shapeType := 13;
      end;
      2:
      begin
        if numValues = 2 then
          shapeType := 5
        else
          shapeType := 15;
      end;
    end;

    FShapeObj.CreateShapeFile(shapeFileName, shapeType)
  end;

  function ExtractXYZ : boolean;
  begin
    RemovePossibleTabsAndDoubleSpaces;
    if PosEx('X', UpperCase(line)) <= 0 then
    begin
      tempList2.Text := AnsiReplaceText(line, ' ', lineEnding);
      x := StrToFloat(TrimRight(TrimLeft(tempList2.Strings[0])));
      y := StrToFloat(TrimRight(TrimLeft(tempList2.Strings[1])));
      if numValues = 3 then
        z := StrToFloat(TrimRight(TrimLeft(tempList2.Strings[2])))
      else
        z := 0;
      Result := true;
    end
    else
      Result := false;
  end;

begin
  OpenDialog1.FileName := '*.txt';
  if not OpenDialog1.Execute then Exit;

  shapeFileName := ChangeFileExt(OpenDialog1.FileName, '');
  if not InputQuery('', 'Give shapefilename', shapeFileName) then Exit;

  if ExistsForPossibleOldShapeFileWithSameName then Exit;

  pointDataType := rgpShapeType.ItemIndex = 0;

  tempList := TStringList.Create;
  tempList2 := TStringList.Create;
  try
    tempList.LoadFromFile(OpenDialog1.FileName);

    // create shapefile
    CreateShapeFile;

    // open shapefile for readwrite
    FShapeObj.Open(shapeFileName, 'rw');

    n := 0;
    while n < tempList.Count do
    begin
      line := TrimRight(TrimLeft(tempList.Strings[n]));
      if (not pointDataType) and ((n = 0) or (not ExtractXYZ)) then
      begin
        if n > 0 then
          FShapeObj.ClosePolygonPolylineShape; // close current shape
        FShapeObj.AddPolygonPolylineShape;
      end
      else
        FShapeObj.Add(x, y, z, 0);

      Inc(n);
    end;

    if not pointDataType then
      FShapeObj.ClosePolygonPolylineShape; // close last shape
    FShapeObj.Close;
  finally
    FreeAndNil(tempList2);
    FreeAndNil(tempList);
  end;
end;

procedure TfrmShapeObjDemo.itemEntityFormatClick(Sender: TObject);
var
  dir, namn : string;
begin
  if not FShapeObj.Select('ro', 'Select shapefile') then Exit;

  dir := FShapeObj.ShapeFileDir;
  namn := ExtractFileName(FShapeObj.ShapeFileName);
  FShapeObj.ExportAsTextFile(dir + namn, dir + namn, -1,
    ProgressBar1, StatusBar1);
end;

procedure TfrmShapeObjDemo.itemSimpleXYZFileClick(Sender: TObject);
var
  dir, namn : string;
  xmin, ymin, xmax, ymax : double;
begin
  if not FShapeObj.Select('ro', 'Select shapefile') then Exit;

  dir := FShapeObj.ShapeFileDir;
  namn := ExtractFileName(FShapeObj.ShapeFileName);
  FShapeObj.ExportAsXYZMTextFile(dir + namn, dir + namn, -1, true,
    FShapeObj.xmin, FShapeObj.ymin, FShapeObj.xmax, FShapeObj.ymax,
    ProgressBar1, StatusBar1);
end;

procedure TfrmShapeObjDemo.lbledPenWidthChange(Sender: TObject);
var
  w, i : integer;
begin
  Val(lbledPenWidth.Text, w, i);
  if i = 0 then
    FPenWidth := w
  else
    lbledPenWidth.Text := IntToStr(FPenWidth);
  DrawShapes;
end;

procedure TfrmShapeObjDemo.PaintBox1MouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
begin
  StatusBar1.SimpleText :=
    'X: ' + FloatToStrF(FWorldViewPortRoutines.wx(x), ffFixed, 15, 2) + ' ' +
    'Y: ' + FloatToStrF(FWorldViewPortRoutines.wy(y), ffFixed, 15, 2);
end;

procedure TfrmShapeObjDemo.itemShowShapeFilesClick(Sender: TObject);
var
  n : integer;
  xmin, ymin, xmax, ymax : single;
begin
  FShapefileList.Clear;
  if not FShapeObj.MultiSelect('Select shapefile(s)', FShapefileList) then Exit;

  // determine total extent
  xmin := 100000000; ymin := 100000000;
  xmax := -100000000; ymax := -100000000;
  n := 0;
  while n < FShapefileList.Count do
  begin
    FShapeObj.Open(FShapefileList.Strings[n], 'ro');
    if xmin > FShapeObj.XMinTot then
      xmin := FShapeObj.XMinTot;
    if ymin > FShapeObj.YMinTot then
      ymin := FShapeObj.YMinTot;
    if xmax < FShapeObj.XMaxTot then
      xmax := FShapeObj.XMaxTot;
    if ymax < FShapeObj.YMaxTot then
      ymax := FShapeObj.YMaxTot;

    Inc(n);
  end;

  xmin := xmin - 50;
  ymin := ymin - 50;
  xmax := xmax + 50;
  ymax := ymax + 50;

  // Set viewport to be able to show world-coordinates
  FWorldViewPortRoutines.SetViewPort(0, 0,
    PaintBox1.ClientWidth, PaintBox1.ClientHeight);
  FWorldViewPortRoutines.SetWorldPort(xmin, ymin, xmax, ymax);

  DrawShapes;
end;

procedure TfrmShapeObjDemo.DrawShapes;
var
    n : integer;
    col : TColor;
begin
  PaintBox1.Canvas.Brush.Color := clWhite;
  PaintBox1.Canvas.FillRect(0, 0, PaintBox1.ClientWidth, PaintBox1.ClientHeight);

  // draw shapes
  n := 0;
  while n < FShapefileList.Count do
  begin
    case n of
      0: col := clRed;
      1: col := clGreen;
      2: col := clBlue;
      3: col := clMaroon;
      4: col := clNavy;
      5: col := clYellow;
    else
      col := clLime;
    end;

    FShapeObj.Open(FShapefileList.Strings[n], 'ro');
    FShapeObj.DrawShapes(PaintBox1.Canvas, FPenWidth,
      col, col, chkgrp1.Checked[0], chkgrp1.Checked[1]);

    Inc(n);
  end;
end;

procedure TfrmShapeObjDemo.itemAttributesClick(Sender: TObject);
begin
  if not FShapeObj.Select('rw', 'Select shapefile') then Exit;
end;

procedure TfrmShapeObjDemo.itemCopyShapeFilesClick(Sender: TObject);
var
  ShapeFileCopyName: string;
begin
  if not FShapeObj.Select('ro', 'Select shapefile') then Exit;

  ShapeFileCopyName :=
    ChangeFileExt(FShapeObj.ShapeFileName, '_Copy');
  if not InputQuery('', 'Give shapefile copyname', ShapeFileCopyName) then Exit;

  FShapeObj.CopyShapeFile(FShapeObj.ShapeFileName, ShapeFileCopyName);
//  CopyShapeFile(ShapeFileName, ShapeFileCopyName: string;
//    DeletedShapesAlso : boolean = false; Prompt : boolean = true): string;
end;

procedure TfrmShapeObjDemo.itemDeleteShapeFilesClick(Sender: TObject);
var
  n : integer;
begin
  FShapefileList.Clear;
  if not FShapeObj.MultiSelect('Select shapefile(s)', FShapefileList) then Exit;

  if FShapeObj.Active then
    FShapeObj.Close;

  n := 0;
  while n < FShapefileList.Count do
  begin
    FShapeObj.DeleteShapeFile(FShapefileList.Strings[n]);

    Inc(n);
  end;
end;

end.

